from .gpt_detector import GPTDetector
from .object_detection import ObjectDetection
from .image_captioning import ImageCaptioning
#from .utils import *

__title__ = "Wyzard"
__version__ = "1.0"
__authors__ = "LyQuid"
__license__ = "MIT License"
__copyright__ = "Copyright 2023-present LyQuid"



import torch
from transformers import AutoImageProcessor, AutoModelForObjectDetection
from PIL import Image
import matplotlib.pyplot as plt
from random import choice


COLORS = ["#ff7f7f", "#ff7fbf", "#ff7fff", "#bf7fff",
            "#7f7fff", "#7fbfff", "#7fffff", "#7fffbf",
            "#7fff7f", "#bfff7f", "#ffff7f", "#ffbf7f"]


class Loader:
    def __init__(self, model, device):
        self.device = device
        self.image_processor = AutoImageProcessor.from_pretrained(model)
        self.model = AutoModelForObjectDetection.from_pretrained(model).to(device)

    def get_figure(self, in_pil_img, in_results, font_size, font_color):
        fdic = {
            "style" : "normal",
            "size" : font_size,
            "color" : font_color,
            "weight" : "bold"
        }
        plt.figure(figsize=(16, 10))
        plt.imshow(in_pil_img)
        ax = plt.gca()

        for score, label, box in zip(in_results["scores"], in_results["labels"], in_results["boxes"]):
            selected_color = choice(COLORS)

            box_int = [i.item() for i in torch.round(box).to(torch.int32)]
            x, y, w, h = box_int[0], box_int[1], box_int[2]-box_int[0], box_int[3]-box_int[1]

            ax.add_patch(plt.Rectangle((x, y), w, h, fill=False, color=selected_color, linewidth=3, alpha=0.8))
            ax.text(x, y, f"{self.model.config.id2label[label.item()]}: {round(score.item()*100, 2)}%", fontdict=fdic, alpha=0.8)

        plt.axis("off")

        return plt.gcf()


    def infer(self, input_img, output_img, threshold, font_size, font_color):
        input_img = Image.open(input_img)
        target_sizes = torch.tensor([input_img.size[::-1]])

        inputs = self.image_processor(images=input_img, return_tensors="pt").to(self.device)
        outputs = self.model(**inputs)

        results = self.image_processor.post_process_object_detection(outputs, threshold=threshold, target_sizes=target_sizes)[0]

        figure = self.get_figure(input_img, results, font_size, font_color)
        figure.savefig(output_img, bbox_inches='tight')

        return results
    
    def models(self):
        return self.model
    

class ObjectDetection:
    def __init__(self, model:str="hustvl/yolos-small", device:str="cuda"):
        self.loader = Loader(model, device)

    def detect(self, input_img, output_img, threshold:int=0.9, font_size:int=18, font_color:str="yellow"):
        return self.loader.infer(input_img, output_img, threshold, font_size, font_color)
    
    def coordinates(self, results):
        coordinates = []
        for score, label, box in zip(results["scores"], results["labels"], results["boxes"]):
            box = [round(i, 2) for i in box.tolist()]
            coordinates.append({"object": self.loader.models().config.id2label[label.item()], "coordinates": box, "precision": round(score.item()*100, 2)})

        return coordinates


from transformers import VisionEncoderDecoderModel, ViTImageProcessor, AutoTokenizer
import torch
from PIL import Image

class ImageCaptioning:
    def __init__(self, model:str="nlpconnect/vit-gpt2-image-captioning", device:str="cuda"):
        self.device = device
        self.model = VisionEncoderDecoderModel.from_pretrained(model).to(device)
        self.feature_extractor = ViTImageProcessor.from_pretrained(model)
        self.tokenizer = AutoTokenizer.from_pretrained(model)

    def predict(self, path:str, max_length:int=16, num_beams:int=4):
        gen_kwargs = {"max_length": max_length, "num_beams": num_beams}
        images = Image.open(path)
        if images.mode != "RGB":
          images = images.convert(mode="RGB")

        pixel_values = self.feature_extractor(images=images, return_tensors="pt").pixel_values
        pixel_values = pixel_values.to(self.device)

        output_ids = self.model.generate(pixel_values, **gen_kwargs)

        preds = self.tokenizer.batch_decode(output_ids, skip_special_tokens=True)
        return preds[0]



class WyzardException(Exception):
    def __init__(self, message: str=None):
        if message is None:
            message = 'An error has occurred within Wyzard'
        super().__init__(message)

class LocalModelNotFound(WyzardException):
    def __init__(self):
        super().__init__("Unable to load local model. Is the model already downloaded and you entered the correct model path?")

class PytorchModelNotFound(WyzardException):
    def __init__(self):
        super().__init__("Unable to find pytorch_model.bin. Did you entered the correct model path?")

class PathDoesntExists(WyzardException):
    def __init__(self, message: str = None):
        super().__init__("Path not found/Doesn't exists.")


from .exceptions import *
import os
import torch
from transformers import AutoModel, AutoTokenizer
from diffusers import StableDiffusionPipeline

def LoadLocalModel(path:str):
    if os.path.exists(path) == True:
        if os.path.exists(os.path.join(path, "pytorch_model.bin")) is True:
            return path
        else:
            raise PytorchModelNotFound()
    else:
        raise LocalModelNotFound()
    

def DownloadModelToLocal(model_name:str, path:str):
    if os.path.exists(path) == True:
        model = AutoModel.from_pretrained(model_name, output_attentions=True)
        tokenizer = AutoTokenizer.from_pretrained(model_name)
        model.save_pretrained(path)
        tokenizer.save_pretrained(path)
        return f"\nModel downloaded to {path}"
    else:
        raise PathDoesntExists()


def DiffusersDownloader(model_name:str, path:str, torch_dtype):
    if os.path.exists(path) == True:
        model = StableDiffusionPipeline.from_pretrained(model_name, torch_dtype=torch_dtype)
        model.save_pretrained(path)
        return f"\nModel downloaded to {path}"
    else:
        raise PathDoesntExists()


"""
Modification of BurhanUlTayyab/GPTZero
Thanks to Burhan Ul tayyab and Nicholas Chua
"""

import torch
import re
from transformers import GPT2LMHeadModel, GPT2TokenizerFast
import time
from .exceptions import *


class Loader:
    def __init__(self, model, device):
        self.device = device
        self.model = model
        self.start_time = time.time()
        self.model = GPT2LMHeadModel.from_pretrained(model).to(device)
        self.tokenizer = GPT2TokenizerFast.from_pretrained(model)

        self.max_length = self.model.config.n_positions
        self.stride = 512
        
    def getResults(self, threshold):
        if threshold < 60:
            label = 0
            return {"output": "The text is likely to be written entirely by a AI.", "code": label}
        elif threshold < 80:
            label = 1
            return {"output": "The text is likely to be written by a combination of Human and AI.", "code":label}
        else:
            label = 2
            return {"output": "The text is likely to be written entirely by a Human.", "code": label}

    def getPPL(self, sentence):
        device = self.device
        encodings = self.tokenizer(sentence, return_tensors="pt").to(device)
        seq_len = encodings.input_ids.size(1)

        nlls = []
        likelihoods = []
        prev_end_loc = 0
        for begin_loc in range(0, seq_len, self.stride):
            end_loc = min(begin_loc + self.max_length, seq_len)
            trg_len = end_loc - prev_end_loc
            input_ids = encodings.input_ids[:, begin_loc:end_loc].to(self.device)
            target_ids = input_ids.clone()
            target_ids[:, :-trg_len] = -100

            with torch.no_grad():
                outputs = self.model(input_ids, labels=target_ids)
                neg_log_likelihood = outputs.loss * trg_len
                likelihoods.append(neg_log_likelihood)

            nlls.append(neg_log_likelihood)

            prev_end_loc = end_loc
            if end_loc == seq_len:
                break
        ppl = int(torch.exp(torch.stack(nlls).sum() / end_loc))
        return ppl
    

    def check(self, sentence):
        total_valid_char = re.findall("[a-zA-Z0-9]+", sentence)
        total_valid_char = sum([len(x) for x in total_valid_char])

        if total_valid_char < 100:
            return {"output": "Please input more text (min 100 characters)"}, "Please input more text (min 100 characters)"
        
        lines = re.split(r'(?<=[.?!][ \[\(])|(?<=\n)\s*',sentence)
        lines = list(filter(lambda x: (x is not None) and (len(x) > 0), lines))

        ppl = self.getPPL(sentence)
        self.perplexity = ppl

        offset = ""
        pp_line = []
        for i, line in enumerate(lines):
            if re.search("[a-zA-Z0-9]+", line) == None:
                continue
            if len(offset) > 0:
                line = offset + line
                offset = ""

            if line[0] == "\n" or line[0] == " ":
                line = line[1:]
            if line[-1] == "\n" or line[-1] == " ":
                line = line[:-1]
            elif line[-1] == "[" or line[-1] == "(":
                offset = line[-1]
                line = line[:-1]
            ppl = self.getPPL(line)
            pp_line.append(ppl)
            
        perplexity_per_line = sum(pp_line)/len(pp_line)
        self.ppl = ppl
        self.burstiness = max(pp_line)
        results = self.getResults(perplexity_per_line)
        self.results = results
        self.process_time = round(time.time() - self.start_time)
        return results
    
    def details(self):
        data = {
            "process_time": self.process_time,
            "output": self.results["output"],
            "perplexity": self.perplexity,
            "perplexity_per_line": self.ppl,
            "burstiness": self.burstiness,
            "code": self.results["code"]
        }

        return data
    


class GPTDetector:
    def __init__(self, model:str="gpt2", device:str="cuda"):
        self.loader = Loader(model=model, device=device)

    def check(self, prompt:str):
        return self.loader.check(prompt)
    
    def details(self):
        return self.loader.details()


from setuptools import setup, find_packages
import re

with open('README.md') as f:
    long_description = f.read()

version = ''
with open('wyzard/__init__.py') as f:
    version = re.search(r'^__version__\s*=\s*[\'"]([^\'"]*)[\'"]', f.read(), re.MULTILINE).group(1)

if not version:
    raise RuntimeError('Version is not set')

def read_requirements():
    with open('requirements.txt', 'r') as req:
        content = req.read()
        requirements = content.split('\n')

    return requirements

setup(
    name='wyzard',
    version=version,
    author='LyQuid',
    author_email='lyquidpersonal@gmail.com',
    description='Run various transformers models from one packages.',
    long_description=long_description,
    long_description_content_type='text/markdown',
    license='MIT License',
    url='https://github.com/LyQuid12/Wyzard',
    project_urls={
        "Source Code": "https://github.com/LyQuid12/Wyzard",
        "Discord": "https://discord.gg/qpT2AeYZRN",
        "Issue tracker": "https://github.com/LyQuid12/Wyzard/issues"
    },
    packages=find_packages(),
    include_package_data=True,
    install_requires=read_requirements(),
    keywords=["python", "machine learning", "ml", "ai", "artificial intelligence", "transformers", "pytorch", "models", "tokenizer"],
    classifiers=[
        #'Development Status :: 5 - Production/Stable',
        "Development Status :: 4 - Beta",
        "Intended Audience :: Developers",
        "Intended Audience :: Science/Research",
        "Topic :: Scientific/Engineering :: Artificial Intelligence",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "License :: OSI Approved :: MIT License",
        "Operating System :: Microsoft :: Windows",
        "Operating System :: MacOS",
        "Operating System :: POSIX :: Linux"
    ]
)
